#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Combined kernel that leverages shared memory for the input vector and uses warp-level primitives for fast reduction
// of the max-pooling results. Each block processes one batch element.

__global__ void combined_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,
    int inF,
    int outF,
    int kernel_size,
    float scale_factor
) {
    int b = blockIdx.x;
    if (b >= B) return;

    // Dynamically allocated shared memory layout:
    // [ shared_x (inF elements) | local_out (outF elements) | warp_sums (blockDim.x/32 elements) ]
    extern __shared__ float smem[];
    float* shared_x = smem;                // size: inF
    float* local_out  = shared_x + inF;     // size: outF
    float* warp_sums  = local_out + outF;    // size: blockDim.x / 32

    // 1. Load the input vector for this batch sample into shared memory
    for (int i = threadIdx.x; i < inF; i += blockDim.x) {
        shared_x[i] = x[b * inF + i];
    }
    __syncthreads();

    // 2. Compute F.linear: each thread computes dot products for output indices in a strided manner
    for (int j = threadIdx.x; j < outF; j += blockDim.x) {
        float sum = bias[j];
        // Use shared_x to reduce global memory traffic
        for (int i = 0; i < inF; i++) {
            sum += shared_x[i] * weight[j * inF + i];
        }
        local_out[j] = sum;
    }
    __syncthreads();

    // 3. Perform max pooling on local_out in segments of size 'kernel_size'
    int pooled_len = (outF >= kernel_size) ? ((outF - kernel_size) / kernel_size + 1) : 0;
    float sum_local = 0.0f;

    // Each thread processes a subset of pooling segments
    for (int seg = threadIdx.x; seg < pooled_len; seg += blockDim.x) {
        int start = seg * kernel_size;
        float max_val = local_out[start];
        for (int k = 1; k < kernel_size; k++) {
            float candidate = local_out[start + k];
            if (candidate > max_val) {
                max_val = candidate;
            }
        }
        sum_local += max_val;
    }

    // 4. Warp-level reduction on sum_local using shuffle intrinsics
    const unsigned int warpSize = 32;
    unsigned int lane = threadIdx.x % warpSize;
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        sum_local += __shfl_down_sync(0xffffffff, sum_local, offset);
    }
    
    // Write the result from each warp's lane 0 to the warp_sums buffer
    if (lane == 0) {
        int warp_id = threadIdx.x / warpSize;
        warp_sums[warp_id] = sum_local;
    }
    __syncthreads();

    // 5. Final reduction: let the first warp reduce the per-warp sums
    int num_warps = blockDim.x / warpSize;
    if (threadIdx.x < num_warps) {
        float final_sum = warp_sums[threadIdx.x];
        for (int offset = num_warps / 2; offset > 0; offset /= 2) {
            final_sum += __shfl_down_sync(0xffffffff, final_sum, offset);
        }
        if (threadIdx.x == 0) {
            output[b] = final_sum * scale_factor;
        }
    }
}


// Host-side forward function callable from Python
at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    const int B = x.size(0);
    const int inF = x.size(1);
    const int outF = weight.size(0);

    auto out = torch::empty({B}, x.options());
    
    const int threads = 256;
    int warps_per_block = threads / 32;
    // Shared memory layout: inF + outF + warps_per_block floats
    size_t sharedMemSize = (inF + outF + warps_per_block) * sizeof(float);

    combined_kernel<<<B, threads, sharedMemSize>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Combined CUDA kernel with shared memory and warp-level reduction optimizations");
}
